# copyright (C) 1997-98 Jean-Luc Fontaine (mailto:jfontain@mygale.org)
# this program is free software: please refer to the BSD type license enclosed in this package

set rcsId {$Id: viewer.tcl,v 1.7 1998/04/30 20:54:34 jfontain Exp $}

class viewer {                                                                             ;# handle viewers related functionalities
    # viewers do not derive from a common interface class but rather support a common set a options through the composite class

    set viewer::(list) {}

    proc viewer {this} {
        new canvasHandles $::canvas -path $widget::($this,path)                                       ;# manage new viewer in canvas
        lappend viewer::(list) $this
    }

    proc ~viewer {this} {
        variable ${this}traces

        foreach array [array names ${this}traces] {                                                       ;# remove all array traces
            trace vdelete ${array}(updates) w "viewer::update $this $array"
        }
        catch {unset ${this}traces}
        ldelete viewer::(list) $this
    }

    virtual proc supportedTypes {this}

    proc view {this cells} {                                                                  ;# cells is a list of data array cells
        foreach cell [parseCells $this $cells] {
            eval monitorCell $this $cell
            set arrays([lindex $cell 0]) {}                                                                        ;# remember array
        }
        foreach array [array names arrays] {                                        ;# update viewer with current values immediately
            update $this $array
        }
    }

    virtual proc monitorCell {this array row column}

    proc parseCells {this cells} {
        if {[llength $cells]==0} return
        set parsed {}
        foreach cell $cells {
            if {![info exists $cell]} {
                error "\"$cell\" cell does not exist"
            }
            parse $cell array row column type
            if {[lsearch -exact [supportedTypes $this] $type]<0} {
                lifoLabel::flash $::messenger "cannot display data of type $type"
                bell
                return
            }
            lappend parsed [list $array $row $column]
        }
        return $parsed
    }

    proc parse {dataCell arrayName rowName columnName typeName} {
        upvar $arrayName array $rowName row $columnName column $typeName type

        if {![regexp {^(.*)\(([0-9]+),([0-9]+)\)$} $dataCell dummy array row column]||($row<0)||($column<0)} {
            error "\"$dataCell\" is not a valid array cell"
        }
        set type [set ${array}($column,type)]
    }

    proc updateInterval {value} {                                             ;# static procedure for updating all viewers intervals
        foreach viewer $viewer::(list) {
            catch {composite::configure $viewer -interval $value}                 ;# some viewers do not support the interval option
        }
    }

    proc label {array row column} {   ;# derive a label from cell using the tabular data index columns. array name must be qualified
        set label {}
        foreach index [set ${array}(indexColumns)] {                                                  ;# use index columns for label
            append label "[set ${array}($row,$index)] "
        }
        append label [set ${array}($column,label)]
        return $label
    }

    virtual proc update {this array args}                                                         ;# update display using cells data

    proc registerTrace {this array} {
        variable ${this}traces

        if {![info exists ${this}traces($array)]} {                                                          ;# if not monitored yet
            trace variable ${array}(updates) w "viewer::update $this $array"                  ;# refresh as module data gets updated
            set ${this}traces($array) 0                                                      ;# initialize number of traces on array
        }
        incr ${this}traces($array)                                                        ;# keep track of number of traces on array
    }

    proc unregisterTrace {this array} {
        variable ${this}traces

        if {[incr ${this}traces($array) -1]<=0} {
            trace vdelete ${array}(updates) w "viewer::update $this $array"                             ;# trace is no longer needed
            unset ${this}traces($array)
        }
    }
}
